/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : mergep.cpp
  DESCRIPTION : merge points in behalf of sorting program (impl)
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920928 AH
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 921213 AH ``amen'' manipulator introduced
  LAST CHANGE : 921112 AH Comment clean up
  LAST CHANGE : 921109 AH TRACE introduced
  LAST CHANGE :
                                                                   */
/*******************************************************************/
#include <assert.h>

#include <fstream.h>
#include "config.h"
#include "lib.h"
#include "mergep.h"

/*****************************************************************************/
/*                                                                           */
/*   FiFoWithMark is a somehwat special FIFO buffer for keeping the merge    */
/*   points, i.e. the points in the intermediate files where the order       */
/*   breaks down. (Implementation is tricky, interface is reasonably         */
/*   general). This implementation restricts the use slightly, because no    */
/*   puts from before the marker can be recovered. But it has the            */
/*   advantage that a single array can be used, where the mergepoints for    */
/*   the file being read and the one being written are stored together.      */
/*   This hinges on the fact that the number of merge points is              */
/*   monotonically decreasing, so getting is faster than putting. We can now */
/*   handle 500Mbyte files in the worst case (1 byte fixed length records).  */
/*                                                                           */
/*****************************************************************************/

// This is done to be able to use static memory, that is
// allocated beforehand, because otherwise all is lost
long FifoWithMark::mergepoint[fifosize];      // Offsets in the new file

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

FifoWithMark::FifoWithMark()
{
  filledput=0;
  filledget=fifosize;
  inuse =fifosize;
};

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

void FifoWithMark::put(long in)
{
   if (filledput>=fifosize)
      cerr<<"Too many records for available memory"<<amen;
   assert(filledput<filledget);
   mergepoint[filledput++] = in;
   TRACE<<"new mergepoint "<<(filledput-1)<<' '<<mergepoint[filledput-1]<<'\n';
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

long *FifoWithMark::content()
{
    assert(filledget>=filledput);
    return &mergepoint[filledget];
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

long FifoWithMark::get()
{
    TRACE<<"got mergepoint "<<(filledget)<<" : "<<mergepoint[filledget]<<'\n';
    assert(filledget<inuse);
    return mergepoint[filledget++];
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

void FifoWithMark::mark()
{
   assert(filledget=inuse);
   TRACE<< "had "<<filledget<<" points, now "<<filledput<<'\n';
   inuse = filledput;
   filledget = 0;
   filledput = 0;
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

int FifoWithMark::leftbeforemark()
{
   return inuse - filledget;
}
