/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : mergep.h
  DESCRIPTION : merge points in behalf of sorting program (def)
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920928 AH
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950206 AH use of constants, cleanup
  LAST CHANGE : 921109 AH comment clean up
  LAST CHANGE :
                                                                   */
/*******************************************************************/

/* Class "First in First out with Marker, shrinking"
 * WARNING : this class may seem rather general, but it has a
 *            "shrinking" property.
 *    Read the restriction carefully if you want to reuse it.
 *
 * Usage is as follows:
 *  call put() to put a long in (first in first out)
 *  call get() to get a long out (first in first out)
 *  call mark() when you want to make a mark at the next long put
 *  call leftbeforemark() to know how many longs are left before the mark.
 *  call content() for a pointer to longs to get.
 *  It is guaranteed that content()[0..leftbeforemark()] is a contiguous array.
 *  The following restrictions apply:
 *  -you cannot get anything anymore that is before the previous mark
 *  -you cannot yet get something put after the current mark
 *  -for every item put after the last mark, you have to got one
 *
 *  NOTE: the last restriction allows us to guarantee the contiguous array
 */

class FifoWithMark
{
private:
   static long FifoWithMark::mergepoint[];      // Offsets in the new file
   int filledput;       // [0..filledput) is the part of mergepoint that
                        //  has been put
   int filledget;       // [filledget..inuse) is the part of mergepoint that
                        //  can be gotten
   int inuse;
public:
   FifoWithMark();
   ~FifoWithMark() { };
   void put(long in) ;  // Put the long into the fifo
   long get();          // The next from the FIFO
   void mark();         // Place marker (between merges)
   long *content();     // Returns long[0] start and long[1] end of mergepiece
   int leftbeforemark();  // The numbers put remaining before the mark
};
