/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : read.cpp
  DESCRIPTION : general interactive input
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920928 AH
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 921213 AH ``amen'' manipulator introduced
  LAST CHANGE : 921112 AH Comment clean up
  LAST CHANGE : 921109 AH TRACE introduced
  LAST CHANGE :
                                                                   */
#include <stdio.h>
#include "config.h"
#include "iostream.h"
#include "lib.h"

FILE *termout = stdout;
FILE *termin = stdin;

void open_term(FILE *out, FILE *in )
{
    termout=out;
    termin=in;
}

void close_term()
{
    fclose(termout);
    fclose(termin);
}

/*********************** readi ********************************************
 *
 * Poses the question "qeustion" on the terminal ans accepts the
 *  answer as an integer.
 * Prints the default integer "deffault" on the terminal.
 * If the operator answers with <CR> only the default is taken.
 * "deffault" may be a NULL pointer. In that case there is no default,
 * and dots are printed instead of a proposed int.
 */

int readi( STRING question, int *deffault )
{
    bool def = (deffault != NULL);
    int i;
    bool success = false;        /* The operator manages to type an int */
    char input[256];

    while ( !success )
    {
        fprintf(termout,"\n");
        fputs(question,termout);
        if ( def )
        {
            fprintf(termout,"\n%d\n",*deffault);
        }
        else
        {
            fprintf(termout,"\n.\n");
        }
        fflush(termout);
        while ( !fgets(input,256,termin) ) {};
        success = ( 1==sscanf(input,"%d",&i) );

        /* Only return --> default */
        if ( '\n'==input[0]  && def )
        {
            i = *deffault;
            success = true;
        }
    }

      return i;
}

/*********************** readb *********************************************
 *
 * Poses the question "question" on the terminal and accepts the
 *  answer as a Boolean, i.e. yes or no answer.
 * Prints the default answer "deffault" on the terminal.
 * If the operator answers with <CR> only, the default is taken.
 * "deffault" may be a NULL pointer. In that case there is no default,
 * and dots are printed instead of a proposed answer.
 */

int readb( STRING question, STRING deffault )
{
    bool def = (deffault != NULL);
    char c;
    bool success = false;        /* The operator manages to type an int */
    char input[256];

    while ( !success )
    {
        fprintf(termout,"\n");
        fputs(question,termout);
        fprintf(termout,"\n%s\n", def? deffault: "..." );
        fflush(termout);
        while ( !fgets(input,256,termin) ) {};

        /* Only return --> default */
        if ( '\n'==input[0]  && def )
            c = deffault[0];
        else
            c = input[0];

        /* Reduce to upper case only */
        if ( c>'Z' ) c -= 0x20 ;

        success = ( 'Y'==c || 'N'==c );
    }

    return 'Y'==c;
}

/*********************** readf *********************************************
 *
 * Poses the question "qeustion" on the terminal and accepts the
 *  answer as a float.
 * Prints the default float "deffault" on the terminal.
 * If the operator answers with <CR> only the default is taken.
 * "deffault" may be a NULL pointer. In that case there is no default,
 * and dots are printed instead of a proposed float.
 */

float readf( STRING question, float *deffault )
{
    bool def = (deffault != NULL);
    float f;
    bool success = false;        /* The operator manages to type a float */
    char input[256];

    while ( !success )
    {
        fprintf(termout,"\n");
        fputs(question,termout);
        if ( def )
            fprintf(termout,"\n%f\n",*deffault);
        else
            fprintf(termout,"\n.\n");
        fflush(termout);
        while ( !fgets(input,256,termin) ) {};
        success = ( 1==sscanf(input,"%f",&f) );

        /* Only return --> default */
        if ( '\n'==input[0]  && def )
        {
            f = *deffault;
            success = true;
        }
    }

    return f;
}

/*********************** reads *********************************************
 *
 * Poses the question "question" on the terminal and accepts the
 *  answer as a string. The trailing <CR> (if any) is removed.
 * Prints the default string "deffault" on the terminal.
 * If the operator answers with <CR> only the default is taken.
 * "deffault" may be a NULL pointer. In that case there is no default,
 * and dots are printed instead of a proposed string.
 * However an empty string is accepted as a default.
 * The string is allocated on the heap.
 */

STRING reads( STRING question, STRING deffault )
{
    bool def = (deffault != NULL);
    char input[256];
    int i;

    fprintf(termout,"\n");
    fputs(question,termout);
    fprintf(termout,"\n%s\n",def? deffault: "..." );
    fflush(termout);
    while ( !fgets(input,256,termin) ) {};
    for (i=1; 0<256; i++)
        if ( input[i]==0 )
        {
            if ( input[i-1]=='\n' ) input[i-1]=0;
            break;
        }


    /* Empty string --> default */
    return ( 0==input[0] && def ?
                 heapc(deffault):
                 heapc(input)
           );
}
