/**************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : scompare.cpp
  DESCRIPTION : installable string comparers for sorting
  PROJECT : SSORT
  AUTOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  WARNING : strcmpl contains dependancy on undocumented Turbo C++ malloc behaviour
  CREATED : 920912 AH
  LAST CHANGE : 950223 AH multiple sort: also initialize weightign
  LAST CHANGE : 950216 AH multiple sort: S is default
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950129 AH len made a pointer
  LAST CHANGE : 941104 AH introduced ssortlib.h
  LAST CHANGE : 941129 AH -I option
  LAST CHANGE : 941125 AH -N option
  LAST CHANGE : 941125 AH Reverse sorting via flag
  LAST CHANGE : 940326 AH -S option now means default sorting
  LAST CHANGE : 921213 AH ``amen'' manipulator introduced
  LAST CHANGE : 921112 AH Comment clean up
  LAST CHANGE : 921110 AH Pointer to comparison hidden
  LAST CHANGE : 921106 AH Introduced CHAR typedef
  LAST CHANGE : 921105 AH Eliminated ``scanf''
  LAST CHANGE : 921105 AH Made a class
                                                                   */
/*******************************************************************/

#include <fstream.h>
#include <strstrea.h>
#include <string.h>
#include <iostream.h>
#include <assert.h>
#include "config.h"
#include "lib.h"
#include "search.h"
#include "sstring.h"
#include "ctype.h"
#include "ssortlib.h"
#include "scompare.h"

/*****************************************************************************/
/*                                                                           */
/*  WARNING: the interaction of different options could lead to a            */
/*  proliferation of different comparisons. For this reason we are always    */
/*  respecting the starting column, be it 1, as if handling the /O option.   */
/*  But the case that is used most often (straight ASCII comparison,         */
/*  ascending order, starting column 1) is singled off. For the same reason  */
/*  all reverse sorting is done with help of a reverse flag, instead of by   */
/*  doubling the number of routines, which would be slightly more efficient. */
/*                                                                           */

/*  WARNING: Comparison as well as table lookup fails in using signed        */
/*  char's. Hence we are using CHAR (=unsigned char).                        */
/*                                                                           */
/*****************************************************************************/

cmpwithfield comparators::installed[installedcomparatorssize];

int comparators::maxcomparator = 0;

//#######################################################################
//##################             ########################################
//################## COMPARATORS ########################################
//##################             ########################################
//#######################################################################


// Comparison starts in this column of a line/record
// Command line option counts from 1, we count from 0
int comparators::startingcol = 0;


/*****************************************************************************/
/*                                                                           */
/*  ``strncmp'' compares field ``field'' of the counted strings ``s1'' and   */
/*  ``s2''. It ignores no characters, so ``startingcol''should be 0. Return  */
/*  value <0 =0 >0 according to if ``s1'' ASCII alfabetically prior, equal   */
/*  or later than ``s2''                                                     */
/*  NOTE: this is the mostly used comparator. In fact it is an optimisation  */
/*  because it is equivalent to ``colstrncmp'' with ``startingcol'' equal    */
/*  zero.                                                                    */
/*                                                                           */
/*****************************************************************************/

int comparators::strncmp( const SortString *s1, const SortString *s2, const int field )
{
   int result;

   int len1 = s1->plen[field+1] - s1->plen[field];
   int len2 = s2->plen[field+1] - s2->plen[field];

   CHAR *pc1 = s1->plen[field]  + s1->start;
   CHAR *pc2 = s2->plen[field]  + s2->start;

   for ( ; len1>0 && len2>0; len1--, len2--  )
      // Warning: uses implicit cast from unsigned char to signed int
      if ( 0 != ( result = *pc1++ - *pc2++ ) ) return result;

    return len1-len2;
}


/*****************************************************************************/
/*                                                                           */
/*  ``colstrncmp'' compares field ``field'' of the counted strings ``s1''    */
/*  and ``s2'' It ignores any characters before ``startingcol''.             */
/*  Return value <0 =0 >0 according to if ``s1'' ASCII alfabetically prior,  */
/*  equal or later than ``s2''                                               */
/*                                                                           */
/*****************************************************************************/

int comparators::colstrncmp( const SortString *s1, const SortString *s2, const int field )
{
   int result;

   int len1 = s1->plen[field+1] - s1->plen[field]-startingcol;
   int len2 = s2->plen[field+1] - s2->plen[field]-startingcol;

   CHAR *pc1 = s1->plen[field]  +  &s1->start[startingcol];
   CHAR *pc2 = s2->plen[field]  +  &s2->start[startingcol];

   for ( ; len1>0 && len2>0; len1--, len2--  )
      // Warning: uses implicit cast from unsigned char to signed int
      if ( 0 != ( result = *pc1++ - *pc2++ ) ) return result;

    return len1-len2;
}


// The following table is used in behalf of sorting while ignoring case.
// Remaining special characters, such with diacritical marks, will be
// treated by ASCII collation, same as ``colstrncmp''
// The following table corresponds with it.
// A character is sorted more towards the end than some other char
//  if its weight is higher than the other character.
// Use the (unsigned) char as an index in ``weightign'' to find this weight.

CHAR comparators::weightign[256];       // Filled in when installing option

/*****************************************************************************/
/*                                                                           */
/*  Initialize the weigth table ``weightign''                                */
/*                                                                           */
/*****************************************************************************/

void comparators::fillweightI( )
{
   for (int i=0; i<256; i++ )
      comparators::weightign[i]=i;
   for (CHAR c='A'; c<='Z'; c++ )
      comparators::weightign[c]= comparators::weightign[tolower(c)];
}


/*****************************************************************************/
/*                                                                           */
/*  ``colstrnicmp'' compares field ``field'' of the counted strings ``s1''   */
/*  and ``s2''. It ignores any characters before ``startingcol''. but a      */
/*  thusly emptied string is sorted towards the end. Return value <0 =0 >0   */
/*  according to if ``s1'' acoording to ASCII collation is prior, equal or   */
/*  later than ``s2'', but ignores difference between upper and lower case.  */
/*  Uses a table to store the weights.                                       */
/*  Precondition: ``fillweightI'' must have been called once.                */
/*                                                                           */
/*****************************************************************************/

int comparators::colstrnicmp( const SortString *s1, const SortString *s2, const int field )
{
   int result;

   int len1 = s1->plen[field+1] - s1->plen[field]-startingcol;
   int len2 = s2->plen[field+1] - s2->plen[field]-startingcol;

   CHAR *pc1 = s1->plen[field]  +  &s1->start[startingcol];
   CHAR *pc2 = s2->plen[field]  +  &s2->start[startingcol];

   for ( ; len1>0 && len2>0; len1--, len2--  )
      // Warning: uses implicit cast from unsigned char to signed int
      if ( 0 != ( result=weightign[*pc1++]-weightign[*pc2++] ) )
         return result;

    return len1-len2;
}


/*****************************************************************************/
/*                                                                           */
/*  ``colstrnlcmp'' compares the counted strings ``s1'' and ``s2'' It        */
/*  ignores any characters before ``startingcol'', but a thusly emptied      */
/*  string is sorted towards the end.                                        */
/*  Return value <0 >0 according to if ``s1'' lexicographically (SORT        */
/*  compatible, observing country characters etc.) prior or later than       */
/*  ``s2''.                                                                  */
/*  If they are equal, we do as if the string with the lowest address is     */
/*  prior. If both are in the same buffer, this makes sorting stable.        */
/*  If both are in different buffers, this trick still makes sorting         */
/*  stable for a file that can be loaded in memory as a whole provided:      */
/*  1. the buffers are ``malloc''ed in ascending order,                      */
/*  2. the compiler allows pointer comparison between buffers, which         */
/*  strictly is illegal, and performs it as expected.                        */
/*  Otherwise an undefined number is returned, but apart from making the     */
/*  sort unstable no harm is done if only the compiler ignores the error.    */
/*  It uses a table for the weights.                                         */
/*                                                                           */
/*****************************************************************************/

// The following table is used for compatibility with the MSDOS
// SORT, provided it is initialised with a call to ``Ssortlib::MSDOSweight()''
// from ``initstrcmpL''.
// A character is sorted more towards the end than some other char
//  if its weight is higher than the other character.
// Use the (unsigned) char as an index in ``weightpage'' to find this weight.
// NOTE: Lexicographic ordering does not just ignore case
//    Instead it has "c small"< "c capital" < "c capital cedille"<
//    "e capital" ..<"e capital aigue"<.."s capital"<"german sz"..

CHAR far *comparators::weightpage;

//
//  Alternative, use your own table:
//  (Remove call to ``Ssortlib::MSDOSweight()'' from ``initstrcmpL'')
//  CHAR comparators::weightpage[] =
//  #include "lextab.h"     // Generated by auxiliary program for US ASCII

int comparators::colstrnlcmp( const SortString *s1, const SortString *s2, const int field)
{
   int result;

   int len1 = s1->plen[field+1] - s1->plen[field]-startingcol;
   int len2 = s2->plen[field+1] - s2->plen[field]-startingcol;

   CHAR *pc1 = s1->plen[field]  +  &s1->start[startingcol];
   CHAR *pc2 = s2->plen[field]  +  &s2->start[startingcol];

   for ( ; len1>0 && len2>0; len1--, len2--  )
      // Warning: uses implicit cast from unsigned char to signed int
      if ( 0 != ( result=weightpage[*pc1++]-weightpage[*pc2++] ) )
         return result;

   //  If one of the string parts has become empty and the other one
   //  has not, than the non-empty part is larger
   if ( (len1<=0) != (len2<=0) ) // Comparison of booleans!
      return len1-len2;

   // The strings are equal, as far as string comparison is concerned.
   // Now force stability by taking into account their order in the file.
   // Not : "return pc1-pc2"
   // That would not work with buffers >32K apart in a compact model.
   // Relies on the fact that buffers are allocated in ascending order!
   // Simply comparing pointers would not do in Borland C++.
   return (CHAR huge*)pc1>(CHAR huge*)pc2 ? 1 : -1 ;
}


/*****************************************************************************/
/*                                                                           */
/*  ``numcolstrncmp'' compares field ``field'' of the counted strings ``s1'' */
/*  and ``s2''                                                               */
/*  It collates all digits from the strings and returns a negative, zero or  */
/*  positive number, respectively if the ``s1'' represents a number smaller  */
/*  than, equal to or larger than ``s2''. Any signs are ignored.             */
/*  It observes the startingcolumn.                                          */
/*                                                                           */
/*****************************************************************************/

int comparators::numcolstrncmp( const SortString *s1, const SortString *s2, const int field )
{
   int result = 0;

   int len1 = s1->plen[field+1] - s1->plen[field]-startingcol; // characters remaining in s1 after pc1
   int len2 = s2->plen[field+1] - s2->plen[field]-startingcol; // similar

   CHAR *pc1 = s1->plen[field]  +  &s1->start[startingcol]; // Where comparison starts
   CHAR *pc2 = s2->plen[field]  +  &s2->start[startingcol];

   // Find first digit in both strings, ignore leading zero's!
   while ( (!isdigit(*pc1) || '0'== *pc1) && len1 > 0 )
   {
      pc1 += 1;
      len1 -= 1;
   }
   while ( (!isdigit(*pc2) || '0'== *pc2) && len2 > 0 )
   {
      pc2 += 1;
      len2 -= 1;
   }

   // Compare digits one by one, until there is a difference.
   while ( len1>0 && len2>0 )
   {
      // Remember m.s. digit
      if ( 0==result )   result = *pc1 - *pc2 ;

      // Find next digits
      while ( --len1>0 && !isdigit(*++pc1) )   {}
      while ( --len2>0 && !isdigit(*++pc2) )   {}
   }

   // One of them has digits left? That one is larger.
   // Both equal lenghth, the m.s. different digit counts
   result = len1>0 || len2>0 ? len1-len2 : result ;
   return result;
}


//#######################################################################
//##################            #########################################
//################## INSTALLERS #########################################
//##################            #########################################
//#######################################################################

/*************************** no option (use ASCII order) *********************/

/*****************************************************************************/
/*                                                                           */
/*   Install the default character comparator.                               */
/*   Also initialisation of ``installed''                                    */
/*                                                                           */
/*****************************************************************************/

void comparators::initdefault( )
{
   maxcomparator = 1;
   installed[0].reverse = false;
   installed[0].field = 0;
   installed[0].method = strncmp;
}

/************* -S option (normal sort) ***************************************/

/*****************************************************************************/
/*                                                                           */
/*  Install the normal character comparator, ASCII collating sequence        */
/*  Precondition: ``installed'' was already initialised.                     */
/*                                                                           */
/*****************************************************************************/

void comparators::initstrcmpS( )
{
   installed[0].method = strncmp;
}


/************* -I option (ignore case ) **************************************/

/*****************************************************************************/
/*                                                                           */
/*   Install the character comparator version that ignores case.             */
/*  Precondition: ``installed'' was already initialised.                     */
/*                                                                           */
/*****************************************************************************/

void comparators::initstrcmpI( )
{
   fillweightI();

   installed[0].method = colstrnicmp;

}


/************* -L option (lexicographic  order) ******************************/

/*****************************************************************************/
/*                                                                           */
/*   Install the lexicographical character comparator.                       */
/*  Precondition: ``installed'' was already initialised.                     */
/*                                                                           */
/*****************************************************************************/

void comparators::initstrcmpL( )
{
   comparators::weightpage = Ssortlib::MSDOSweight();

   installed[0].method = colstrnlcmp;
}


/************* -N option (numeric order) *************************************/

/*****************************************************************************/
/*                                                                           */
/*   Install the numeric comparator.                                         */
/*  Precondition: ``installed'' was already initialised.                     */
/*                                                                           */
/*****************************************************************************/

void comparators::initstrcmpN( )
{
   installed[0].method = numcolstrncmp;
}


/********************** R option (reverse order) *****************************/

/*****************************************************************************/
/*                                                                           */
/*   Reverse the character comparator.                                       */
/*  Precondition: ``installed'' was already initialised.                     */
/*                                                                           */
/*****************************************************************************/

void comparators::initstrcmpR( )
{
   installed[0].reverse = true;
}

/******************** -O option (starting column) ****************************/

/*****************************************************************************/
/*                                                                           */
/*   Change the starting column to the argument in ``s''                     */
/*                                                                           */
/*****************************************************************************/

void comparators::initstartingcol( STRING s )
{
   istrstream strstr( s, strlen(s) );

   strstr >> startingcol;
   if ( strstr.fail() )
      cerr<<"Want column specified in decimal :"<<s<<amen;

   startingcol -= 1;

   if (startingcol<0)
      cerr<<"Column specified should be 1 (default) or more: "<<(startingcol+1)<<amen;

   // Take care that the installed method can handle columns
   if ( installed[0].method == strncmp )
      installed[0].method = colstrncmp;
}


/********************** -M option (multiple fields) **************************/

/*****************************************************************************/
/*                                                                           */
/*  Install the comparisions specified with the -M option into the array     */
/*  ``installed''.                                                           */
/*  ``specification'' is a sequence of                                       */
/*  - field numbers                                                          */
/*  - a sort specifier, one of                                               */
/*      I: case insensitive sort,                                            */
/*      L: lexicographic sort,                                               */
/*      N: numeric sort,                                                     */
/*      S: normal sort                                                       */
/*  - optional : R indicating reverse sorting                                */
/*                                                                           */
/*****************************************************************************/

void comparators::initstrcmpM( STRING specification )
{
   char *pc = specification;

   // May be needed, for ``colstrnicmp'' resp. ''colstrnlcmp''
   fillweightI();                                            
   comparators::weightpage = Ssortlib::MSDOSweight();

   for ( maxcomparator=0; *pc; maxcomparator++)
   {
      if ( maxcomparator >= installedcomparatorssize )
         cerr<<"Maximum number of comparisons handled is: "
            <<maxcomparator<<amen;

      cmpwithfield *pi = & installed[maxcomparator];
      pi->reverse = false;      // Default
      pi->method = strncmp;     // Default

      // Interpret digits
      if ( !isdigit(*pc) )
         cerr<< "No digit were expected in -M option: "<<specification<<amen;
      pi->field = *pc++ - '0';
      while ( isdigit(*pc) )
      {  // Multiply by 10 and add new digit
         pi->field *= 10;
         pi->field += *pc++ - '0';
      }

      if ( 0 == *pc )
               cerr<< "No letter were expected in -M option: "<<specification<<amen;
      // Interpret letters
      while ( *pc && !isdigit(*pc) )
         switch ( toupper(*pc++) )
         {
            case 'S' : pi->method = strncmp;        break;
            case 'L' : pi->method = colstrnlcmp;    break;
            case 'I' : pi->method = colstrnicmp;    break;
            case 'N' : pi->method = numcolstrncmp;  break;
            case 'R' : pi->reverse = true;          break;
            default :
               cerr<< "Invalid letter in -M option: "<<specification<<amen;
         }
   }
}
