/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : scompare.h
  DESCRIPTION : installable string comparers for sorting
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920912 AH
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 941129 AH -I option
  LAST CHANGE : 941125 AH -N option (not tested)
  LAST CHANGE : 941125 AH Reverse sorting via flag
  LAST CHANGE : 921112 AH Pointer to comparison hidden
  LAST CHANGE : 921109 AH comment clean up
  LAST CHANGE : 921105 AH Made a class
                                                                   */
/*******************************************************************/

/*****************************************************************************/
/*                                                                           */
/*   The class ``comparators'' has only static members. It hides the         */
/*   different ways to compare SortString's from the outside world.          */
/*   In the outside view there are calls to configure this class, with       */
/*   prefix ``init'', and ``stringcompare'' that is what this class is       */
/*   all about, the comparison method actually selected.                     */
/*   Each of the methods may have additional configuration data such as      */
/*   the record length for fixed record length.                              */
/*   Installation is intended to be done at the start of the program once,   */
/*   other usage of this module has not been tested. There is, however, no   */
/*   reason why changing the method during program runs should not have the  */
/*   expected results.                                                       */
/*   Names are about counted strings, not zero ended strings. Hence the      */
/*   names are based on ``strncmp'' instead of ``strcmp''.                   */
/*   The 'l' in ``strnlcmp'' is of lexicographic.                            */
/*                                                                           */
/*   Note:                                                                   */
/*   It might seem appropriate that these methods are instantiations of      */
/*   a virtual stringcompare, but there I can find no way in C++ to have     */
/*   this result in a reasonable structured program. Strousup points out     */
/*   that techniques using pointers to subroutines have still their          */
/*   place in C++ to provide "polymorphic" routines (C++ sec. ed.            */
/*   4.6.9). Probably this is a good example of this.                        */
/*                                                                           */
/*****************************************************************************/

// The general type of a comparer,
// compare the fields ``field'' of ``s1'' and ``s2'' in some way.
typedef int cmp( const SortString *s1, const SortString *s2, const int field);

// Here the field to which the method should be applied is known too.
typedef struct {                            // Installed comparators
       cmp *method;                           // to be used in this order
       bool reverse;                          // Reverse order?
       int field;                             // Which field
   } cmpwithfield;

class comparators
{
private:
   static int startingcol;
   static cmp strncmp;
   static cmp colstrncmp;
   static cmp colstrnlcmp;
   static cmp colstrnicmp;
   static cmp numcolstrncmp;
   static cmp *pstringcompare ;
   static CHAR far *comparators::weightpage;  // Weights for MSDOS collate
   static CHAR comparators::weightign[256] ;  // Weights to case insensitive
   static void comparators::fillweightI( );   // Get those weights
   static cmpwithfield installed[];           // Installed comparators
   static int maxcomparator;                  // Number of comparators installed

public:
   // Initialisers, used from the main program
   static void initdefault( );   // To be called allways once at the start
   static void initstrcmpS( );                 // -S option: default
   static void initstrcmpI( );                 // -I option
   static void initstrcmpL( );                 // -L option
   static void initstrcmpN( );                 // -N option
   static void initstrcmpR( );                 // -R option
   static void initstrcmpM( STRING s );        // -M option
   static void initstartingcol( STRING s );    // -O option (/O <column>)

   // The crux: this class serves to initialize ``stringcompare''
   // Return a number <0 =0 >0 whether s1<s1 s1=s2 s1>s2
   inline static int stringcompare( const SortString *s1, const SortString *s2)
   {
      int result;
      for (int i=0; i<maxcomparator; i++)
      {
         if ( installed[i].reverse )
            result = -(*installed[i].method)( s1, s2, installed[i].field );
         else
            result = (*installed[i].method)( s1, s2, installed[i].field );

         if ( result ) return result;
      }
      return result;
   };
};
