/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : search.cpp
  DESCRIPTION : installable record separator finders for sorting
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920906 AH
  LAST CHANGE : 950216 AH separator instead of terminator allowed
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950205 AH explicit expression for CSYNC allowed
  LAST CHANGE : 950126 AH CSYNC implemented
  LAST CHANGE : 950126 AH CANY can be '\n'
  LAST CHANGE : 950126 AH Uniform treatment of escapes: \b .. \0F
  LAST CHANGE : 941204 AH streams clean up
  LAST CHANGE : 941129 AH seppositionE made an array, '@' introduced
  LAST CHANGE : 941125 AH cleanup regular expressions
  LAST CHANGE : 940326 AH -S option removed
  LAST CHANGE : 940324 AH OFFSET nonsense & version removed
  LAST CHANGE : 940320 AH Made separatorE an int array
  LAST CHANGE : 940320 AH cleanup ``ismemberofsetE''
  LAST CHANGE : 940320 AH '|' replaced by '$', '^' use improved
  LAST CHANGE : 921215 AH cleanup regular expressions
  LAST CHANGE : 921213 AH ``amen'' manipulator introduced
  LAST CHANGE : 921212 AH option -C accepts '\t' and such
  LAST CHANGE : 921126 AH pfindseparator -> findseparator
  LAST CHANGE : 921126 AH cclass -> ismemberofset
  LAST CHANGE : 921113 AH -S gave infinite loop
  LAST CHANGE : 921111 AH Comment clean up
  LAST CHANGE : 921106 AH Introduced CHAR typedef
  LAST CHANGE : 921105 AH Made a class
  LAST CHANGE : 921105 AH Eliminated ``scanf''
  LAST CHANGE : 920924 AH removed MSDOS text cludge
                                                                   */
/*******************************************************************/

#include <string.h>
#include <ctype.h>
#include <strstrea.h>
#include <iomanip.h>
#include <assert.h>
#include "config.h"
#include "lib.h"
#include "search.h"

/*******************************************************************/

/*
   WARNING: We are using CHAR (=unsigned char). (See scompare.cpp).
            This result in warnings at calls of library functions.
 */

/************* static data members *************************/

/* The separator finder that is installed using the various routines
   of this module. Default is separation by lines.
 */
CHAR * (*Separators::findseparator )( CHAR *, int ) =
     Separators::findseparatorC;

int Separators::recordlengthestimate = 30;  // estimate for storage allocation

/************* option -C (single character) ****************/

CHAR Separators::separatorC = '\n';     // Actual separator used

bool Separators::hasbeenfixedC = false;

/*****************************************************************************/
/*                                                                           */
/*  ``fixableC'' tries to fix the character buffer ``buf[0..len-1] '' by     */
/*  adding a separator. This is only possible if records are separated by    */
/*  characters. In that case the separator is added to the content of the    */
/*  buffer ``buf'' and the bufferlength ``len'' is adjusted. Returns "the    */
/*  record was fixed".                                                       */
/*                                                                           */
/*****************************************************************************/

bool Separators::fixableC( char *buf, int &len )
{
   hasbeenfixedC = findseparator == findseparatorC ;

   if ( hasbeenfixedC )
      buf[len++] = separatorC;

   return hasbeenfixedC ;
}


/*****************************************************************************/
/*                                                                           */
/*  ``removeseparator'' looks whether a separator has been added to the      */
/*  last record. If so, the separator of the last record is removed from     */
/*  the file with name ``filename'' that resulted from sorting.              */
/*  ``binaryfile'' indicates that it is not a text file. If it does be,      */
/*  and the separator was '\n', two characters have to be removed instead    */
/*  of one.                                                                  */
/*                                                                           */
/*****************************************************************************/

void Separators::removeseparatorC( char *filename, bool binaryfile )
{
   if ( hasbeenfixedC )
   {
      int correction =
                      binaryfile         ? -1 :
                      separatorC == '\n' ? -2 :
                                           -1    ;

      changefilesize( filename, correction );
   }
}


/*****************************************************************************/
/*                                                                           */
/*   Return the start of the next record from buf[0..len].                   */
/*   Records are supposed to end with the ``separatorC'' that should be      */
/*   installed at the same time as this method.                              */
/*   Returns 0 if no match found.                                            */
/*                                                                           */
/*****************************************************************************/

CHAR * Separators::findseparatorC( CHAR *buf, int len )
{
   for ( CHAR *pc = buf; len--; pc++ )
      if ( *pc == separatorC ) return pc+1;
   return 0;
}


/*****************************************************************************/
/*                                                                           */
/*   Install the single character separator. Interpret ``s'' as either a     */
/*   single character (including the escaped \b \f \n \t ) or a hex          */
/*   denotation of a character, and install it as the separator.             */
/*                                                                           */
/*****************************************************************************/

void Separators::initseparatorC( STRING s )
{
   STRING pc = s;

   separatorC = nextchar( pc );

   if ( *pc )
      cerr<<"Specify one character as separator for C : "<<s<<amen;

   findseparator = findseparatorC;
}

/*****************************************************************************/
/*                                                                           */
/*   Return the next character at ``sp'' taking into account escapes         */
/*   starting with '\'. This may be single character escapes, followed by a  */
/*   lower case letter: \b \f \n \t. Or it can be a double character escapes */
/*   with two hex digits, only uppercase A..F accepted.                      */
/*   ``sp'' is left pointing after the letters consumed.                     */
/*                                                                           */
/*****************************************************************************/

CHAR Separators::nextchar( STRING &sp )
{
   CHAR c = *sp++;
   CHAR * spcopy = sp;  // Remember for error messages only

   // Normal character return
   if ( c!='\\' )
      return c;

   if ((c = *sp++) == '\000')
      cerr<<"Command line has an empty \\-escape : "<<amen;

   // Single character escape
   switch (c)
   {
      case 'b': return '\b';
      case 'f': return '\f';
      case 'n': return '\n';
      case 'r': return '\r';
      case 't': return '\t';
      case '\\': return '\\';
   }

   CHAR * hexdigits = "0123456789ABCDEF";
   CHAR * digit0 = strchr( hexdigits, c );

   if ( !digit0 )
   {
      sp -= 1;
      return '\\'; // Leave escape handling to caller
   }

   if ( *sp == '\000')
      cerr<<"Command line has a short \\-escape : "<<spcopy<<amen;

   CHAR * digit1 = strchr( hexdigits, *sp++ );

   if ( !digit1 )
      cerr<<"Command line has a wrong \\-escape : "<<spcopy<<amen;

   return (digit0-hexdigits)*16 + (digit1-hexdigits);
}


/************* option -E (regular Expression) ************************/

// In the expression '$' is used to denoted the end of the record, e.g.
// "\n\n\n$[^\n]" means a record ends in two empty lines, but the
// following line must not be empty.
// A consequence is that a real '$' must be escaped.
// So escaped are '\\','\$','\t' and '\n'

/*****************************************************************************/
/*                                                                           */
/*   Note: SeparatorE[n] may contain either a character, or one of the       */
/*   explicit enumerations following, or a closure (CCHAR|CLOS), or a        */
/*   count. This depends from the order.                                     */
/*   Instead of using a union, we make it an integer area and rely as        */
/*   much as possible on safe conversions.                                   */
/*                                                                           */
/*****************************************************************************/

// Auxiliary string: compiled expression.
// Contains a mix of ``charclass'' enums, counts and ``CHAR''.
// All three types are subtypes of ``int''.
int *Separators::patternE;
int Separators::lenpatternE;  // Its allocated length, used for safety

// Points at where the next fields start in the buffer
// Stored outside of recursively called routines
// seppositionE[0] field that starts at '^' or before the first '$'
// seppositionE[i] field that begin with i-th '$'
// seppositionE[maxsep] start of next record,
// lines up next time with seppositionE[0]
CHAR **Separators::seppositionE;        // Allocated array of CHAR *
// ``maxfield'' is the length of this array
// it indicates where the one and only record starts and ends
int  Separators::maxfield=1;
// Indicates the field that has been passed by findseparatorE
// the last time it was called.
int  Separators::fieldpointer;

// The end of the buffer where a match is to be found.
// As yet we use a 0 to mark the end of the buffer.
CHAR *Separators::limitE = NULL;

enum charclass {
   CCHAR = 2  ,    // Specific character
   CANY  = 4  ,    // Any character
   CSET  = 6  ,    // A character set
   CSETC = 8  ,    // A character set, characterized by its complement
   CLOS  = 01 ,    // And this in to get the closure of one of the above
   CEOF  = 10 ,    // Marks the separation between fields
   CSYNC = 12 ,    // Match as few characters as possible
   CEND  = 0       // Marks the end of the regular expression
};

// Layout of expression in ``patternE''.
// NOTE: THIS IS A SEQUENTIAL LAYOUT!
// It consists of the following elements (separated by ',' | is bitwise or )
// CCHAR c , CLOS|CCHAR c , CANY , CLOS|CANY , CEOF #field ,
// CSET count (count-1)*c , CSETC count (count-1)*c ,
// CLOS|CSET count (count-1)*c , CLOS|CSETC count (count-1)*c ,
// CSYNC , CEND
// ``c'' means any character except null ('\000'). Null is not allowed,
// because it must not be present in the expression to be matched, it is
// needed as a terminator. ``count'' and ``#field'' are integers.
// #field 0 matches the beginning of the record (say '^') and 1 ..
// matches the field spearators ('$' of implied in '@'). The last field
// separator is the CEOF 0 of the next record.
// CEOF and CEND are restricted
// Note: the position in ``patternE'' determines whether it is a
// ``charclass'', a count or a character. If sequentially parsed this
// does not lead to ambiguity despite overlapping ranges
// (All unsigned char's are acceptable in the expression as c.)


/*****************************************************************************/
/*                                                                           */
/*  Compile the string ``s'' into ``patternE' with the tricky format         */
/*  described above. This is in order to be able to swiftly parse text       */
/*  searching for regular expressions.                                       */
/*                                                                           */
/*****************************************************************************/

void Separators::compileE(STRING s )
{
   CHAR c;

   // Pointer into compiled string
   int* ep = patternE;
   // Pointers into source string
   STRING sp = s;        // ``nextchar'' fails if declared CHAR *

   int *lastep;   // Points to a character class that could be closed
   // Counts the members of a class (between [ ])  == #members +1 (!!)
   int cclcnt;

   // Initialize the separator counter
   maxfield = 0;

   // The '^' at the beginning has a special meaning (see manual)
   if ( '^'!=s[0] )  *ep++ = CSYNC;

   while (c = nextchar(sp))
   {
      assert( ep<patternE+lenpatternE );

      switch (c)
      {
      case '.':
         lastep = ep;   // Candidate for closure, leads to CSYNC
         *ep++ = CANY;
         break;

      case '*':
         if (!lastep) cerr<<"No expression for closure : "<<s<<amen;

         if ( CANY != *lastep )
            *lastep |= CLOS;
         else
            *lastep = CSYNC;    // Special case

         lastep = NULL;         // Case closed
         break;

      case '$':
         lastep = NULL;         // Do not allow e.g. A$*
         *ep++ = CEOF;
         *ep++ = ++maxfield;
         break;

      case '^':
         if ( sp != s+1 )
            cerr<<"'^' only allowed at start of expression:"<<s<<amen;
         lastep = NULL;         // Do not allow e.g. A^*
         *ep++ = CEOF;
         *ep++ = 0;
         break;

      case '[':
         lastep = ep;   // Candidate for closure
         *ep++ = *sp=='^'? CSETC: CSET;   // May later be CLOS-ed
         *ep++ = 0;                     // Going to contain class-count
         if ( *sp=='^' ) sp += 1;  // Consume '^': negator, not a character

         // Put characters in set, at least one. So do{ }while.
         c = nextchar(sp);
         cclcnt = 1;    // One extra count!
         do {
            *ep++ = c;
            cclcnt++;
            assert( ep<patternE+lenpatternE );
         } while ((c = nextchar(sp)) != ']');
         lastep[1] = cclcnt;
         break;

      case '@':
         c = nextchar(sp);
         lastep = NULL;        // No closure possible
         *ep++ = CSETC|CLOS;   // A complement set
         *ep++ = 2;            // With one character excluded:
         *ep++ = c;            //   the field separator
         *ep++ = CCHAR;        // Followed by a single char:
         *ep++ = c;            //   the field separator again
         *ep++ = CEOF;         // Mark the field separation point
         *ep++ = ++maxfield;
         break;

      case '\\':
         c = *sp++;            // Do not use nextchar! No nested esapes!
         if (c=='\000' ) cerr<<"Expect a character after escape: "<<s<<amen;
         // Treat as normal character : fall through to default

      default:
         lastep = ep;   // Candidate for closure
         *ep++ = CCHAR;
         *ep++ = c;
         break;
      }
   }

   // Add default field marker if none found as yet
   if ( 0==maxfield )
   {
      *ep++ = CEOF;
      *ep++ = ++maxfield;
   }

   // End of input string: close compiled expression
   assert( ep<patternE+lenpatternE );
   *ep++ = CEND;
}


#ifdef DEBUG

/*****************************************************************************/
/*                                                                           */
/*   ``PrintSingleCharE'' prints the single character ``c''. It need not be  */
/*   a printable character, but it may not be '\000'                         */
/*                                                                           */
/*****************************************************************************/

void Separators::PrintSingleCharE(CHAR c )
{
   switch( c )
   {
      case '\*'       : cout<<"'\\*' ";         break;
      case '\@'       : cout<<"'\\@' ";         break;
      case '\$'       : cout<<"'\\$' ";         break;
      case '\b'       : cout<<"'\\b' ";         break;
      case '\f'       : cout<<"'\\f' ";         break;
      case '\n'       : cout<<"'\\n' ";         break;
      case '\t'       : cout<<"'\\t' ";         break;
      default         : if ( '\000' != c )
                           cout<<"'"<<c<<"' ";
                        else
                           cout<<"No character where expected\n";
                        break;
   }
}


/*****************************************************************************/
/*                                                                           */
/*  Decipher the string in ``patternE '' to reveal the tricky format partly  */
/*  described above. This is a debugging function.                           */
/*                                                                           */
/*****************************************************************************/

void Separators::decipherE( void )
{
   int *sp = Separators:: patternE;
   CHAR c;
   int i;
   bool tobeclosed;

   // Print hex versions of characters
   while ( *sp || CEOF==sp[-1]) cout<<setw(2)<<setfill('0')<<hex<<*sp++<<' ';

   // Print interpreted versions of chars
   cout<<"   ";
   sp = patternE;

   while ( CEND != *sp )
   {
      tobeclosed = false;
      switch( c= *sp++ )        // Note all the fall throughs!
      {
      case CEOF       : cout<<(*sp++?" $ ":" ^ ");
                        break;

      case CCHAR|CLOS : tobeclosed = true;
      case CCHAR      : PrintSingleCharE( c= *sp++);
                        break;

      case CANY|CLOS  : tobeclosed = true;
      case CANY       : cout<<" . ";
                        break;

      case CSYNC      : cout<<" ... ";
                        break;

      case CSETC|CLOS : tobeclosed = true;
      case CSETC      : cout<<"~";
                        goto set;

      case CSET|CLOS  : tobeclosed = true;
      case CSET       : goto set;

      set             : cout<<"[";
                        i = *sp++ -1;
                        if (i<=0) cout<<"illegal length in set\n";
                        while ( i-- )  PrintSingleCharE( c= *sp++);
                        cout<<"]";
                        break;

      default         : cout<<"wrong format\n";
                        break;
      }
      if ( tobeclosed ) cout<<"* ";
   }
   cout<<"\n";
}
#endif


/*****************************************************************************/
/*                                                                           */
/*  Returns "``ac'' is a member of the set ``set''".                         */
/*  ``set'' is a part of the compiled expression representing a "class"      */
/*  of characters, i.e. a set between '[' and ']'.                           */
/*  Structure of ``set'' : count, (count-1) characters.                      */
/*                                                                           */
/*****************************************************************************/

bool Separators::ismemberofsetE(int *set ,  CHAR c )
{
   int remaining = *set++ -1;

   while (remaining)
      if (remaining >= 3 && set[1] == '-')
      {
         if (c >= set[0] && c <= set[2] ) return true;
         set += 3;
         remaining -= 3;
      }
      else
      {
         if (*set++ == c) return true;
         remaining -= 1;
      }

   return false;
}


/*****************************************************************************/
/*                                                                           */
/*  Returns "the string at ``lp'' matches the whole of the regular           */
/*  expression ``ep''"                                                       */
/*                                                                           */
/*****************************************************************************/

bool Separators::matchE( CHAR *lp ,  int *ep )
{
   CHAR *closstart;     // Position in ``lp'' where current closure starts

   // Each pass through this loop handles an element of the compiled
   // expression (where ``ep'' points), including its closure (if any).
   // Returns as soon as a mismatch is detected.
   for (;;)
   {
      switch (charclass(*ep++))
      {
      case CEND:
         return true;

      case CEOF:
         seppositionE[*ep++] = lp;
         break;

      case CCHAR:
         if (*ep != *lp) return false;
         ep++, lp++;
         break;

      case CCHAR|CLOS:
         closstart = lp;
         while (*lp == *ep)  lp++;
         ep++;
         goto star;

      case CANY:
         if ( *lp==0 ) return false;
         lp++;
         break;

      case CSYNC:
         while ( lp<limitE )
         {
            bool result;
            if (result = matchE(lp, ep)) return result;
            lp += 1;
         }
         return false;
         break;

      case CSET:
         if (!ismemberofsetE(ep, *lp++)) return false;
         ep += *ep;
         break;

      case CSET|CLOS:
         closstart = lp;
         while ( ismemberofsetE(ep, *lp) ) lp += 1;
         ep += *ep;
         goto star;

      case CSETC:
         if ( *lp==0 ) return false;
         if (ismemberofsetE(ep, *lp++)) return false;
         ep += *ep;
         break;

      case CSETC|CLOS:
         closstart = lp;
         while (*lp && !ismemberofsetE(ep, *lp) ) lp += 1;
         ep += *ep;
         goto star;

      star:
         // `lp' points to the first non-matching character
         // At this point we have matched as much patterns as possible
         // Backtrack if we have matched too much
         do {
            bool result;
            if (result = matchE(lp, ep)) return result;
            lp -= 1;
         } while ( lp >= closstart);     // Backtrack if no match
         return false;

      default:
         cerr<<"Error in compiled expression at :"<<(ep-patternE)<<amen;
      }
   }
}


/*****************************************************************************/
/*                                                                           */
/*   Return the start of the next record from buf[0..len-1].                 */
/*   Records are supposed to end with the first match with the regular       */
/*   expression that should have been compiled into `patternE' at the        */
/*   installation time of this method. The precise separation of records is  */
/*   where the expression contains the separation marker '$'.                */
/*   Returns 0 if no match found. This may be caused by the buffer being     */
/*   exhausted. The caller may decide to refill the buffer and go on. There  */
/*  is a problem, if there could have been a match with more, e.g. a pattern */
/*  ending in "\n*$" or any star. The caller can not easily detect this,     */
/*  because it gets the fields served one at a time but it has too.          */
/*                                                                           */
/*****************************************************************************/

CHAR * Separators::findseparatorE(CHAR *buf ,  int len )
{
   limitE = buf+len;

   if ( fieldpointer==maxfield ) fieldpointer = 0;
   if ( 0==fieldpointer )
   {
      seppositionE[maxfield] = NULL;   // Reset

      // Consider the pattern matched if
      //    1. matchE returned TRUE          or
      //    2. all fields matched and no context present after that
      if ( !matchE( buf, patternE ) &&
            seppositionE[maxfield] != buf+len )
         return 0;
   }
   return seppositionE[++fieldpointer];
}


/*****************************************************************************/
/*                                                                           */
/*   Install the expression separator finder.                                */
/*   Interpret ``s'' as an regular expression and install it as the          */
/*   separator.                                                              */
/*                                                                           */
/*****************************************************************************/

void Separators::initseparatorE(STRING s )
{
   /* Compile the searchpattern into the expression buffer. */
   lenpatternE = 10 + 4*strlen(s) ;    // Estimate!
   patternE = new int[ lenpatternE ];
   compileE( s );
   seppositionE = new CHAR *[ maxfield ];
   findseparator = findseparatorE;
   fieldpointer = 0;
}


/************* option -F (fixed record length) ************************/

int Separators::cumfieldlengthF[fieldsinrecordsize];  // Fixed field lengths

/*****************************************************************************/
/*                                                                           */
/*  Return the start of the next field from buf[0..len]. Records are         */
/*  supposed to have layout defined by ``cumfieldlengthF''. Returns 0 if no  */
/*  full record is present anymore, otherwise the next field. It remembers   */
/*  which field it is in. Because the offsets of the fields are the same     */
/*  for each record, ``maxfield'' consequitive calls convey essentially the  */
/*  same information.                                                        */
/*                                                                           */
/*****************************************************************************/

CHAR * Separators::findseparatorF( CHAR *buf, int len )
{
   if ( fieldpointer==maxfield ) fieldpointer = 0;
   if ( 0==fieldpointer )
   {
      if ( len < cumfieldlengthF[maxfield] )
         return 0;
   }
   return buf + cumfieldlengthF[++fieldpointer];
}


/*****************************************************************************/
/*                                                                           */
/*  Install the fixed length separator finder. ``s'' should contain the      */
/*  lengths of the fields, separated by ':'. Install into                    */
/*  ``cumfieldlengthF' the accumulated lengths of the fields.                */
/*  Consequently the last entry contains the total record length.            */
/*                                                                           */
/*****************************************************************************/

void Separators::initseparatorF( STRING specification )
{
   findseparator = findseparatorF;
   fieldpointer = 0;

   char *pc = specification;
   int cumfield = 0;
   maxfield = 0;

   cumfieldlengthF[0]=cumfield;
   while ( *pc )
   {
      // Interpret digits
      if ( !isdigit(*pc) )
         cerr<< "No digit were expected in -F option: "<<specification<<amen;
      int fieldlen = *pc++ - '0';
      while ( isdigit(*pc) )
      {  // Multiply by 10 and add new digit
         fieldlen *= 10;
         fieldlen += *pc++ - '0';
      }
      cumfield += fieldlen;
      cumfieldlengthF[++maxfield]=cumfield;
      if ( 0 != *pc )        // Do not increment if already at end of string!
         if ( ':' != *pc++ )
            cerr<< "No colon were expected in -F option: "<<specification<<amen;
   }

   recordlengthestimate = cumfield;  // Estimate for sure
}
