/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : search.h
  DESCRIPTION : installable record separator finders for sorting
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920911 AH
  LAST CHANGE : 950216 AH separator instead of terminator allowed
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950126 AH Uniform treatment of escapes: \b .. \0F
  LAST CHANGE : 941129 AH seppositionE made an array, '@' introduced
  LAST CHANGE : 941125 AH cleanup regular expressions
  LAST CHANGE : 940326 AH -S option now means default sorting
  LAST CHANGE : 940320 AH Made patternE an int array
  LAST CHANGE : 940320 AH cleanup ``ismemberofsetE'' and some unused var's
  LAST CHANGE : 940320 AH '|' replaced by '$', '^' use improved
  LAST CHANGE : 921127 AH killed wrap around findseparator
  LAST CHANGE : 921126 AH cclass -> ismemberofsetE, DEBUG: decipher
  LAST CHANGE : 921111 AH Comment clean up
  LAST CHANGE : 921105 AH Made a class
                                                                   */
/*******************************************************************/

/*****************************************************************************/
/*                                                                           */
/*   The class Separators contains 3 selectable record separators :          */
/*   findseparatorX, each with some extra configuration parameters and       */
/*   possibly auxiliary routines. At each point in time one of those has     */
/*   been selected using initseparatorX. Selection means that the            */
/*   reference to subroutine ``Separators::pfindseparator'' is filled in     */
/*   and subsequently used to separate records in class Buffer.              */
/*   Installation is intended to be done at the start of the program once,   */
/*   other usage of this module has not been tested. There is, however, no   */
/*   reason why changing the method during program runs should not have the  */
/*   expected results.                                                       */
/*   See also the note on subroutine pointers in SCOMPARE.H                  */
/*                                                                           */
/*****************************************************************************/

class Separators
{
#ifdef DEBUG
public:
#else
private:
#endif
   static CHAR nextchar( STRING & );      // Handles escapes
   static int fieldpointer;           // The progress in returning fields

   static CHAR *findseparatorC( CHAR *, int );
   static CHAR separatorC;

   static CHAR *findseparatorE( CHAR *, int );
   static int *patternE;
   static int lenpatternE;
   static CHAR **seppositionE;
   static CHAR *limitE;
   static bool matchE( CHAR *alp, int *aep );
   static void compileE( STRING astr );
   static bool ismemberofsetE(  int *aset, CHAR ac );
#ifdef DEBUG
   static void decipherE( void );   // Show regular expression in debug format
   static void PrintSingleCharE( CHAR ); // Auxiliary for ``decipher''
#endif

   static CHAR *findseparatorF( CHAR *, int );
   static int cumfieldlengthF[fieldsinrecordsize]; // cumulative field lengths
   static bool hasbeenfixedC;   // A record terminator was added.

public:
   static int maxfield;         // The number of fields in a record

   static void initseparatorC( STRING s );  // Install character terminator
   static bool Separators::fixableC( char *buf, int& len );
                                            // Last record is fixable
   static void removeseparatorC( char *filename, bool binaryfile );
                                            // Unfix last record
   static void initseparatorE( STRING s );  // Install regular expr separator
   static void initseparatorF( STRING s );  // Install fixed record length

   // Return the start of the next record from buf[0..n] using installed method
   static CHAR * (*findseparator )( CHAR *, int );
   static int recordlengthestimate;  // estimate for storage allocation
};
