/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE: ssort.cpp
  DESCRIPTION : main of sorting program
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920904 AH
  LAST CHANGE : 950310 AH cosmetic in behalf of release
  LAST CHANGE : 950221 AH Little memory for streambuffer
  LAST CHANGE : 950220 AH Working with little memory
  LAST CHANGE : 950216 AH separator instead of separator allowed
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950126 AH -B option ate extra argument, -C default not bin
  LAST CHANGE : 941129 AH -I option
  LAST CHANGE : 941125 AH -N option
  LAST CHANGE : 941125 Options with no arg's where buggy.
  LAST CHANGE : 940326 AH -S option now means default sorting
  LAST CHANGE : 940326 AH -B option introduced
  LAST CHANGE : 940324 AH -+ option made -O option, optimization correct
  LAST CHANGE : 940323 AH Message on write errors
  LAST CHANGE : 921214 AH destructor for buffer in ``streamcopy''
  LAST CHANGE : 921213 AH ``amen'' manipulator introduced
  LAST CHANGE : 921111 AH Comment clean up
  LAST CHANGE : 921110 AH bug: unlinking of output file failed
  LAST CHANGE : 921109 AH TRACE introduced
  LAST CHANGE : 921108 AH Just one merge in ``merge''
  LAST CHANGE : 921108 AH End bug for merging (1!=leftover)
  LAST CHANGE : 921108 AH Cleanup swapping overflowheap and sortedheap
  LAST CHANGE : 921107 AH Intoduced FifOWithMark
  LAST CHANGE : 921106 AH Splitted off some subroutines
  LAST CHANGE : 921105 AH Made ``Separators'' class
  LAST CHANGE : 921105 AH Made ``comparators'' class
  LAST CHANGE : 920916 cin and cout possible
                                                                   */
/*******************************************************************/

                      char hlp1[] =
"                 SSORT: SUPER SORT UTILITY\n"
"                       VERSION: ";/* Version goes here */
                      char hlp2[] =
"\n"
"           COPYRIGHT: 1992@2000 by SP&AR&C B.V. HOLLAND\n"
"           >>  USAGE SUBJECT TO GNU PUBLIC LICENSE  <<\n"
"                      SEE SSORT.TXT\n\n"
" SSORT [-{H|?|R|B}] [-{I|L|N|S}] [-{C|F|E} SEP] [-O COL] [-M SPEC] [IN [OUT]]\n"
" sorts lines from IN-file to OUT-file\n"
" If OUT is missing, sorts to standard out;\n"
" If IN is also missing, sorts from standard in to standard out.\n"
" Options are case insensitive. No redirection is possible for binary files.\n"
" Options :\n"
"      -H -? this help    -B binary file     -R reverse sorting\n"
"      -S normal sort, ASCII values ordering (default)\n"
"      -I normal sort, but ignore case\n"
"      -L use lexicographical ordering: compatible with SORT.EXE\n"
"      -N use numeric ordering: ignore non-digits\n"
"      -O COL start comparison at offset COL: compatible with SORT.EXE\n"
"      -C SEP is a character separator, no fields, text file\n"
"      -F SEP gives the field lengths, binary\n"
"      -E SEP define fields by a regular expression, text file\n"
"      -M SPEC multiple sorting, each field treated separately\n";

// Excerpt to go with fatal error messages :
char usage[] = "\nUsage:\n"
" SSORT [-{H|?|R|B}] [-{I|L|N|S}] [-{C|F|E} SEP] [-O COL] [-M SPEC] [IN [OUT]]\n";

/*
 *  THE FOLLOWING IS STILL PROBLEMATIC
 *  If the last record or the first record is incomplete such that
 *  record boundaries mix up during reordering this is detected, but
 *  continuing with the sort is probably not reasonable.
 */

/*   The lines marked with "OPT" are optimizations.                       */
/*   They may be left out without difference except for speed.            */

/*******************************************************************/

#include <fstream.h>
#include <process.h>    // for exit()
#include <stdio.h>
#include <alloc.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>

#include "config.h"
#include "lib.h"
#include "search.h"
#include "sstring.h"
#include "scompare.h"
#include "sheap.h"
#include "mergep.h"
#include "sbuffer.h"
#include "version.h"

/*****************************************************************************/
/*                                                                           */
/*  ``help'' prints the help message. Help is not executed in                */
/*  combination with real work.                                              */
/*                                                                           */
/*****************************************************************************/

void help()
{
   cout <<  hlp1 << version << hlp2;
}


/*****************************************************************************/
/*                                                                           */
/*  ``binaryfile'' means "The file is considered binary". Its default:       */
/*  false i.e. text.                                                         */
/*  Precise definition of what text mode means:                              */
/*   1. all trailing ^Z are supposed not to belong to the file               */
/*   2. '\r' '\n' pairs are turned into a single '\n' on input and           */
/*      this is restored on output.                                          */
/*  Redirected input or output (``cin'' and ``cout'') cannot be binary.      */
/*  This is not our fault. All intermediate files are binary (ios::binary),  */
/*  because we want to use tellp() and seekg().                              */
/*                                                                           */
/*****************************************************************************/

bool binaryfile = false;

/*****************************************************************************/
/*                                                                           */
/*   Copy from a stream ``in'' to a stream ``out''.                          */
/*   While writing out in text mode, an '\r' is inserted before each '\n'.   */
/*   This is intentional and happens for ``cout'' as well as in outputfiles  */
/*   opened in text mode. Hopefully they are reinserted at places the user   */
/*   finds reasonable. But this is beyond our control.                       */
/*                                                                           */
/*****************************************************************************/

PRIVATE void streamcopy ( istream& in, ostream& out )
{
   int wanted = streamcopybufsize;
   char *buffer;

   while ( ! (buffer = new char[wanted]) )
      wanted = wanted*3L/4L;

   TRACE<<"Converting from binary using a buffer of length "<<wanted<<'\n';
   while( !in.eof() )
   {
      in.read( buffer, wanted );
      out.write( buffer, in.gcount() );
   }
   if ( !out.good() ) cerr<<"Write error, probably disk full"<<amen;
   delete [] buffer;
}


/*****************************************************************************/
/*                                                                           */
/*   Write all SortString's of all *pSortBuffer's from ``sortedheap'' to     */
/*   the (already open) destination stream ``dest''. The result is sorted.   */
/*                                                                           */
/*****************************************************************************/

PRIVATE void writeout (
  SortedHeap< pSortBuffer > &sortedheap,
  ostream& dest
)
{
   while( !sortedheap.empty() )
   {
      pSortBuffer psb = sortedheap.removeitem();
OPT   while ( !psb.sb->empty() && (sortedheap.empty()||sortedheap>=psb) )
         dest << psb.sb->removeitem();
      if ( !psb.sb->empty() )
         sortedheap.additem(psb);
      else
         delete psb.sb;
   }
   if ( !dest.good() ) cerr<<"Write error, probably disk full"<<amen;
}


/*****************************************************************************/
/*                                                                           */
/*   ``sortedheap'' has been filled from some source, but that source is     */
/*   not yet exhausted. ``overflowheap'' is empty initially.                 */
/*   ``makeintermediate'' will write an intermediate file SCRATCH2 with as   */
/*   large sorted chunks as possible. All buffers from ``sortedheap'' are    */
/*   emptied and result in one large sorted chunk. The buffers from          */
/*   ``sortedheap'' are refilled during this process, but they are moved     */
/*   from ``sortedheap'' to ``overflowheap''. When ``sortedheap'' becomes    */
/*   empty the roles of ``overflowheap'' and ``sortedheap'' are              */
/*   interchanged. This process has no calls to new or delete.               */
/*   The points where the order breaks down are stored in ``mergepoint'',    */
/*   with a marker placed after all those points. (See ``FifoWithMark'')     */
/*   Note: We really want pointers here, no references, otherwise swapping   */
/*   interferes with destruction.                                            */
/*   The lines marked with "OPT" are optimizations. They increase            */
/*   speed dramatically on sorted and almost sorted files.                   */
/*                                                                           */
/*****************************************************************************/

void makeintermediate(
   FifoWithMark &mergepoint,
   SortedHeap< pSortBuffer > *sortedheap,
   SortedHeap< pSortBuffer > *overflowheap
)
{
   ofstream dest;

   TRACE<<"Making an intermediate file\n";
   SortedHeap< pSortBuffer > *swap;

   dest.open( SCRATCH2, ios::binary );  // All intermediate files binary
   assert(dest);

   // Write out full load of buffers
   // and remember mergepoints, i.e. points in the output file where the
   // order breaks down.
   do {   // until everything read
      mergepoint.put( dest.tellp() );
      while( !sortedheap->empty() )
      {
         pSortBuffer psb = sortedheap->removeitem();
         // Optimization : Remove all possible records from psb.sb
         //                not just one
OPT      while ( !psb.sb->empty() && (sortedheap->empty()||*sortedheap>=psb) )
            dest << psb.sb->removeitem();
         if ( !psb.sb->empty() )
            sortedheap->additem(psb);
         else if ( psb.sb->refill(), !psb.sb->empty() )
OPT      // Optimization : Put back in sortedheap whenever possible
OPT         if ( !sortedheap->empty() && !(*sortedheap>psb) )
OPT            sortedheap->additem(psb);
OPT         else
               overflowheap->additem(psb);
         else
            delete psb.sb;
      }

      swap = overflowheap;
      overflowheap = sortedheap;
      sortedheap = swap;
   } while( !sortedheap->empty() );

   if ( !dest.good() ) cerr<<"Write error, probably disk full"<<amen;
   mergepoint.put( dest.tellp() );
   dest.close();
   mergepoint.mark();
}


/*****************************************************************************/
/*                                                                           */
/*   Pre and post-conditions are that the SCRATCH2 output file contains a    */
/*   (at least) partially sorted result, with mergepoints in                 */
/*   ``mergepoint''.                                                         */
/*   ``merge'' creates a new SCRATCH2 with less mergepoints (in              */
/*   ``mergepoint''). Result is totally sorted iff ``mergepoint'' has two    */
/*   points left.                                                            */
/*   The lines marked with "OPT" are optimizations. They increase            */
/*   speed but not so dramatically as with ``makeintermediate''.             */
/*                                                                           */
/*****************************************************************************/

void merge( FifoWithMark &mergepoint )
{
   ifstream source;     // declare input and output streams
   ofstream dest;

   SortedHeap< pMergeBuffer > sortedheap(heapsize);

   // Copy input to output, merging with as many buffers as possible
   unlink( SCRATCH1 );
   rename( SCRATCH2, SCRATCH1 );
   source.open( SCRATCH1, ios::nocreate|ios::binary );
   assert(source);
   dest.open( SCRATCH2, ios::binary );
   assert(dest);

   int wanted = normalbufsize;          // What we want for a buffersize

   // If we can't get what we want, try to want what we can get
   // Need two buffers for merging!
   while ( !SortBuffer::enough_memory_for( wanted, 2 ) )
        wanted = wanted*3L/4L;         // 25% less

   do {
      // Initialise full load of buffers into sortedheap
      pMergeBuffer psb;
      do {
         // Get a new mergebuffer at the current mergepoint
         psb.sb = new MergeBuffer(source, mergepoint.content(), wanted );
         mergepoint.get();   // next mergepoint
         sortedheap.additem(psb);
      } while ( 1!=mergepoint.leftbeforemark() && psb.sb->enough_memory() );

      TRACE<<"Merging "<<sortedheap.nlines<<"-ways\n";

      if ( sortedheap.nlines<2 && 1!=mergepoint.leftbeforemark() )
         cerr<<"not enough memory for merging"<<amen;

      mergepoint.put( dest.tellp() );
      do{
         // Merge and write and refill
         pMergeBuffer psb = sortedheap.removeitem();
OPT      while ( !psb.sb->empty() && ( sortedheap.empty() || sortedheap>=psb ) )
            dest<<psb.sb->removeitem();
         if ( !psb.sb->empty() )
            sortedheap.additem(psb);
         else if ( psb.sb->refill(), !psb.sb->empty() )
            sortedheap.additem(psb);
         else
            delete psb.sb;
      } while ( !sortedheap.empty() );

   } while ( 1!=mergepoint.leftbeforemark() );

   if ( !dest.good() ) cerr<<"Write error, probably disk full"<<amen;
   mergepoint.put( dest.tellp() );
   dest.close();

   mergepoint.mark();

   if ( !source.good() ) cerr<<"Read error, cause unknown"<<amen;
   source.close();
   unlink( SCRATCH1 );
}


/*****************************************************************************/
/*                                                                           */
/*   The program's behaviour is adjusted according to the option ``option''  */
/*   with optional argument ``arg''. Often these subroutines install         */
/*   pointers to subroutines for e.g. record limit finding or string         */
/*   comparison. Alphabetic order.                                           */
/*                                                                           */
/*****************************************************************************/

void handle_option( char option, STRING arg )
{
   switch( option )
   {
   case 'B' :
      binaryfile = true;
      break;

   case 'C' :
      if ( !arg ) cerr<<"Specify argument for option "<<option<<amen;
      Separators::initseparatorC( arg );
      // Rely on default for ``binaryfile'' to be false.
      break;

   case 'E' :
      if ( !arg ) cerr<<"Specify argument for option "<<option<<amen;
      Separators::initseparatorE( arg );
      // Rely on default for ``binaryfile'' to be false.
      break;

   case 'F' :
      if ( !arg ) cerr<<"Specify argument for option "<<option<<amen;
      binaryfile = true;
      Separators::initseparatorF( arg );
      break;

   case 'H' :
   case '?' :
      help();
      exit(0);
      break;

   case 'I' :
      comparators::initstrcmpI();
      break;

   case 'L' :
      comparators::initstrcmpL();
      break;

   case 'N' :
      comparators::initstrcmpN();
      break;

   case 'M' :
      if ( !arg ) cerr<<"Specify argument for option "<<option<<amen;
      comparators::initstrcmpM( arg );
      break;

   case 'O' :
      if ( !arg ) cerr<<"Specify argument for option "<<option<<amen;
      comparators::initstartingcol( arg );
      break;

   case 'R' :
      comparators::initstrcmpR();
      break;

   case 'S' :
      comparators::initstrcmpS();
      break;

   default  : cerr<<"Wrong option"<<amen;
   }
}


/*****************************************************************************/
/*                                                                           */
/*   Guess what? The main program, actually.                                 */
/*                                                                           */
/*****************************************************************************/

void main(int argc, char* argv[])  // access command-line arguments
{
   STRING arg;          // The argument currently under investigation
   STRING inputfile = 0;    // The input file name (if any)
   STRING outputfile =0;  // The output file name (if any)

   ifstream source;     // declare input and output streams
   ofstream dest;
   int nfiles=0;        // Number of file arguments

   // Default configuration, mostly superfluous and for reference only
   // Essential: before handling arguments!
   comparators::initdefault();
///binaryfile = false;
///Separators::initseparatorC( "\n" );
///comparators::initstartingcol("1");

   // Interpret arguments
   while ( arg =*++argv ) // Pop next argument
   {
      if ( arg[0] == '-' || arg[0] == '/' )
      {
          char option = toupper(arg[1]);

          // Some options need an argument :
          if ( strchr( "BILNRS", option ) )
             arg = 0;           // No argument
          else if ( arg[2]=='\0' )
             arg = *++argv;     // Use next word
          else
             arg = &arg[2];     // Use remainder of word, possibly ignored

          handle_option( option, arg );
      }
      else if ( nfiles==0 )
      {
         inputfile = arg;
         nfiles += 1;
      }
      else if ( nfiles==1 )
      {
         outputfile = arg;
         nfiles += 1;
      }
      else
          cerr<<"Too many file spec's : "<<arg<<amen;
   }

   if ( nfiles>=1 ) // Otherwise ``cin'' is going to be used
   {
      source.open( inputfile, ios::nocreate | (binaryfile?ios::binary:0) );
      if ( !source ) cerr<<"Could not open input file : "<<inputfile<<amen;
   }

   if ( nfiles>=2 ) // Otherwise ``cout'' is going to be used
   {
      dest.open( outputfile, binaryfile?ios::binary:0 );
      if ( !dest ) cerr<<"Could not create output file : "<<outputfile<<amen;
   }

   // Allocation up front, see warning
   SortedHeap< pSortBuffer > sortedheap(heapsize);
   SortedHeap< pSortBuffer > overflowheap(heapsize);
   FifoWithMark merges;      // Offsets in the new file

   /*
   WARNING: in the following part MergeBuffers and SortBuffers
   are allocated and deleted. It is important that no allocation
   to the heap interferes. Otherwise fragmentation will kill merging
   of large files.
   */

   // Get full load of buffers
   pSortBuffer psb;
   SortBuffer::remainderstart=0;        // No left overs from previous reads
   int wanted = normalbufsize;          // What we want for a buffersize

   // If we can't get what we want, try to want what we can get
   while ( !SortBuffer::enough_memory_for( wanted ) )
        wanted = wanted*3L/4L;         // 25% less

   do {
      psb.sb = new SortBuffer(
                 nfiles>=1? (istream &)source: (istream &)cin,
                 wanted
               );
      sortedheap.additem(psb);
   } while ( !psb.sb->eof &&
             psb.sb->enough_memory() &&
             !sortedheap.full()      );

   if ( psb.sb->eof )       // Normal case: whole file fits in memory
   {
      writeout( sortedheap, nfiles>=2? (ostream&)dest: (ostream&)cout );
   }
   else
   {
      // Create an intermediate file in SCRATCH2

      // Whether ``cin'' is being used is hidden in ``sortedheap''
      // Do not close ``source'' as long as ``sortedheap'' is in use!
      makeintermediate( merges, &sortedheap, &overflowheap );

      // Keep on merging until one piece
      while( merges.leftbeforemark()>2 ) merge(merges);

      // The output is now available in binary format in file SCRATCH2
OPT   if ( binaryfile &&  // SCRATCH2 already in proper format
OPT        outputfile[1]!=':' &&       // Output file not on other disk
OPT        !strchr(outputfile,'\\') && //  or in other
OPT        !strchr(outputfile,'/')     //  directory
OPT      )
OPT   {  // Just rename SCRATCH2
OPT      dest.close();  // Otherwise unlink would fail
OPT      nfiles -= 1;   // One less file to close, a bit dirty
OPT      unlink(outputfile);
OPT      rename( SCRATCH2, outputfile);
OPT   }
OPT   else  // Convert to text or to proper location or to ``cout''
      {
         ifstream scratch;
         scratch.open( SCRATCH2, ios::nocreate|ios::binary  );
         streamcopy( scratch, nfiles>=2? (ostream&)dest: (ostream&)cout );
         scratch.close();
         unlink( SCRATCH2 );
      }
   }

   if ( nfiles>=1 ) source.close(); // Otherwise ``cin'' was used
   if ( nfiles>=2 ) dest.close();   // Or ``cout'' was used, or allready closed

   // A missing separator may have been added,
   // give ``Separators'' a chance to remove it again.
   if ( outputfile )
      Separators::removeseparatorC( outputfile, binaryfile );

   exit(0);
}
